from pygame import draw
from engine.Gfx import gfx
import copy

class Entity:
	"""
	This is a single entity in the world possibly made up of several animated frames.
	These should be exported from the puffer vector editor.
	"""
	def __init__(self, world):
		"""
		All we need is the world that we belong to.
		"""
		self.world = world
		
		self.frames = []
		self.animations = {"": [[]]}
		self.frame = 0.0
		self.animation = ""
		self.framerate = 1.0
		
		self.width = 0
		self.scale = 1
		self.worldscale = 1.0
		self.position = [0, 0]
		self.lastposition = [0, 0]
		self.scaledposition = [0, 0]
	
	def AddAnimation(self, name, frames, actions=[]):
		"""
		Add a puffer animation to this entity's available animations.
		"""
		# make a copy of the polygon data incase someone else wants the original
		new = copy.deepcopy(frames)
		# run the transforms we want on the polygon data
		[a(new) for a in actions]
		self.animations[name] = new
	
	def Animate(self, which):
		"""
		Run one of my animations (increasing framerate each time we're called)
		"""
		if which != self.animation:
			self.animation = which
			self.frame = 0
		else:
			self.IncFrame()
	
	def IncFrame(self):
		"""
		Increase the frame by one.
		"""
		self.frame = (self.frame + self.framerate) % len(self.animations[self.animation])
	
	def Normalise(self, frames):
		"""
		Figure out the width and height of a set of frames and then normalise all points within them.
		Normalises with the values of the size of the puffer editor screen.
		"""
		low = [10000, 10000]
		high = [0, 0]
		for f in frames:
			for p in f:
				for points in p['points']:
					low[0] = min(low[0], points[0])
					low[1] = min(low[1], points[1])
					high[0] = max(high[0], points[0])
					high[1] = max(high[1], points[1])
		
		# normalise frames using the max and min values we found and the built in screensize of the puffer editor
		remove = []
		for f in frames:
			if len(f):
				for p in f:
					p['points'] = [[(x - (high[0] - low[0]) / 2.0 - low[0]) / 400.0, (y - (high[1] - low[1]) - low[1]) / 400.0] for x,y in p['points']]
				f.reverse()
			else:
				remove.append(f)
		
		self.width = (high[0] - low[0]) / 400.0
		
		# remove any frames that we found that were empty
		for f in remove:
			frames.remove(f)
		
		return frames
	
	def Resize(self, frames):
		"""
		Figure out the width and height of a set of frames and then normalise all points within them.
		Like normalise but it uses the size of the thing, not it's editor size.
		"""
		low = [60000, 60000]
		high = [-60000, -60000]
		for f in frames:
			for p in f:
				for points in p['points']:
					low[0] = min(low[0], points[0])
					low[1] = min(low[1], points[1])
					high[0] = max(high[0], points[0])
					high[1] = max(high[1], points[1])
		
		# normalise frames using the max and min values we found and the built in screensize of the puffer editor
		remove = []
		for f in frames:
			if len(f):
				for p in f:
					p['points'] = [[(x - (high[0] - low[0]) / 2.0 - low[0]) / (high[1] - low[1]), (y - (high[1] - low[1]) - low[1]) / (high[1] - low[1])] for x,y in p['points']]
				f.reverse()
			else:
				remove.append(f)
		
		self.width = (high[0] - low[0]) / (high[1] - low[1])
		
		# remove any frames that we found that were empty
		for f in remove:
			frames.remove(f)
		
		return frames
	
	def ReplaceColor(self, a, b):
		"""
		Replace all instances of a certain color within a puffer model with another color/
		"""
		for anm in self.animations:
			for f in self.animations[anm]:
				for p in f:
					if p['color'] == a:
						p['color'] = b
					if p['outline'] == a:
						p['outline'] = b
	
	def Darken(self, c, amount=20):
		"""
		Darken a color.
		"""
		return [max(0, p - amount) for p in c]
	
	def Reverse(self, frames):
		"""
		Flips frames horizontally.
		"""
		for f in frames:
			for p in f:
				p['points'] = [[-x, y] for x,y in p['points']]
		
		return frames

	def Update(self):
		"""
		Do whatever you have to every frame. (override me higher up, probably to pump events)
		"""
	
	def Draw(self):
		"""
		Render this entity onto a surface.
		"""
		if self.worldscale > 0:
			for p in self.animations[self.animation][int(self.frame)]:
				points = [[(self.scale * self.worldscale * x + self.scaledposition[0]) * gfx.screen.get_width() + gfx.screen.get_width() / 2, self.scale * self.worldscale * y * gfx.screen.get_width() + self.scaledposition[1] * gfx.screen.get_height() + gfx.screen.get_height() / 2] for x,y in p['points']]
				if len(points) > 1:
					if p['filled'] and len(points) > 2:
						draw.polygon(gfx.screen, p['color'], points, 0)
					draw.lines(gfx.screen, p['outline'], p['closed'], points, p['width'])
	
	def Rescale(self):
		"""
		Recalculate the scaling parameters for when the entity moves
		"""
		self.worldscale, self.scaledposition = self.world.Scale(self.position[0], self.position[1])
	
	def Move(self, which, velocity):
		"""
		Move the entity to the left at the speed of it's velocity
		"""
		self.position[which] = self.position[which] + velocity
		self.Rescale()

